/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Copyright (C) 2006 Michael Brown <mbrown@fensystems.co.uk>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <config/general.h>

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL )

	.arch i386

/* Image compression enabled */
#define COMPRESS 1

#define MAX_EBDA_KBYTES 0xff
//#define MAX_EBDA_KBYTES 0x3f

//#define EBDA_DEBUG
//#define EBDA_HEXDUMP

/* Protected mode flag */
#define CR0_PE 1

/*****************************************************************************
 * Utility function: print character (with LF -> LF,CR translation)
 *
 * Parameters:
 *   %al : character to print
 *   %ds:di : output buffer (or %di=0 to print to console)
 * Returns:
 *   %ds:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_character
print_character:
	/* Preserve registers */
	pushw	%ax
	pushw	%bx
	pushw	%bp
	/* If %di is non-zero, write character to buffer and exit */
	testw	%di, %di
	jz	1f
	movb	%al, %ds:(%di)
	incw	%di
	jmp	3f
1:	/* Print character */
	movw	$0x0007, %bx		/* page 0, attribute 7 (normal) */
	movb	$0x0e, %ah		/* write char, tty mode */
	cmpb	$0x0a, %al		/* '\n'? */
	jne	2f
	int	$0x10
	movb	$0x0d, %al
2:	int	$0x10
	/* Restore registers and return */
3:	popw	%bp
	popw	%bx
	popw	%ax
	ret
	.size	print_character, . - print_character

/*****************************************************************************
 * Utility function: print a NUL-terminated string
 *
 * Parameters:
 *   %ds:si : string to print
 *   %ds:di : output buffer (or %di=0 to print to console)
 * Returns:
 *   %ds:si : character after terminating NUL
 *   %ds:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_message
print_message:
	/* Preserve registers */
	pushw	%ax
	/* Print string */
1: 	lodsb
	testb	%al, %al
	je	2f
	call	print_character
	jmp	1b
2:	/* Restore registers and return */
	popw	%ax
	ret
	.size	print_message, . - print_message

/*****************************************************************************
 * Utility functions: print hex digit/byte/word/dword
 *
 * Parameters:
 *   %al (low nibble) : digit to print
 *   %al : byte to print
 *   %ax : word to print
 *   %eax : dword to print
 *   %ds:di : output buffer (or %di=0 to print to console)
 * Returns:
 *   %ds:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_hex_dword
print_hex_dword:
	rorl	$16, %eax
	call	print_hex_word
	rorl	$16, %eax
	/* Fall through */
	.size	print_hex_dword, . - print_hex_dword
	.globl	print_hex_word
print_hex_word:
	xchgb	%al, %ah
	call	print_hex_byte
	xchgb	%al, %ah
	/* Fall through */
	.size	print_hex_word, . - print_hex_word
	.globl	print_hex_byte
print_hex_byte:
	rorb	$4, %al
	call	print_hex_nibble
	rorb	$4, %al
	/* Fall through */
	.size	print_hex_byte, . - print_hex_byte
	.globl	print_hex_nibble
print_hex_nibble:
	/* Preserve registers */
	pushw	%ax
	/* Print digit (technique by Norbert Juffa <norbert.juffa@amd.com> */
	andb	$0x0f, %al
	cmpb	$10, %al
	sbbb	$0x69, %al
	das
	call	print_character
	/* Restore registers and return */
	popw	%ax
	ret
	.size	print_hex_nibble, . - print_hex_nibble

/*****************************************************************************
 * Utility function: print PCI bus:dev.fn
 *
 * Parameters:
 *   %ax : PCI bus:dev.fn to print
 *   %ds:di : output buffer (or %di=0 to print to console)
 * Returns:
 *   %ds:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_pci_busdevfn
print_pci_busdevfn:
	/* Preserve registers */
	pushw	%ax
	/* Print bus */
	xchgb	%al, %ah
	call	print_hex_byte
	/* Print ":" */
	movb	$':', %al
	call	print_character
	/* Print device */
	movb	%ah, %al
	shrb	$3, %al
	call	print_hex_byte
	/* Print "." */
	movb	$'.', %al
	call	print_character
	/* Print function */
	movb	%ah, %al
	andb	$0x07, %al
	call	print_hex_nibble
	/* Restore registers and return */
	popw	%ax
	ret
	.size	print_pci_busdevfn, . - print_pci_busdevfn

#ifdef BCV

/*****************************************************************************
 * Utility function: print mac address
 *
 * Parameters:
 *   %es:si : source buffer network byte ordered
 *   %ds:di : output buffer (or %di=0 to print to console)
 *   NOTE! the segments are the wrong way round. That's intentional
 * Returns:
 *   %ds:si : next byte of buffer
 *   %es:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_mac_addr
print_mac_addr:
	/* Preserve registers */
	pushw	%cx
	pushw	%ax
	movw	$6, %cx
1:	movb	%es:0(%si), %al
	call	print_hex_byte
	inc	%si
	loop	1b
	/* Restore registers and return */
	popw	%ax
	popw	%cx
	ret
	.size	print_mac_addr, . - print_mac_addr

#endif	/* BCV */

/*****************************************************************************
 * Utility function: clear current line
 *
 * Parameters:
 *   %ds:di : output buffer (or %di=0 to print to console)
 * Returns:
 *   %ds:di : next character in output buffer (if applicable)
 *****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	print_kill_line
print_kill_line:
	/* Preserve registers */
	pushw	%ax
	pushw	%cx
	/* Print CR */
	movb	$'\r', %al
	call	print_character
	/* Print 79 spaces */
	movb	$' ', %al
	movw	$79, %cx
1:	call	print_character
	loop	1b
	/* Print CR */
	movb	$'\r', %al
	call	print_character
	/* Restore registers and return */
	popw	%cx
	popw	%ax
	ret
	.size	print_kill_line, . - print_kill_line

/****************************************************************************
 * copy_bytes
 *
 * Copy bytes
 *
 * Parameters:
 *   %ds:esi : source address
 *   %es:edi : destination address
 *   %ecx : length
 * Returns:
 *   %ds:esi : next source address
 *   %es:edi : next destination address
 * Corrupts:
 *   None
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
copy_bytes:
	pushl	%ecx
	rep addr32 movsb
	popl	%ecx
	ret
	.size	copy_bytes, . - copy_bytes

/****************************************************************************
 * zero_bytes
 *
 * Zero bytes
 *
 * Parameters:
 *   %es:edi : destination address
 *   %ecx : length
 * Returns:
 *   %es:edi : next destination address
 * Corrupts:
 *   None
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
zero_bytes:
	pushl	%ecx
	pushw	%ax
	xorw	%ax, %ax
	rep addr32 stosb
	popw	%ax
	popl	%ecx
	ret
	.size	zero_bytes, . - zero_bytes

/****************************************************************************
 * process_bytes
 *
 * Call memcpy()-like function
 *
 * Parameters:
 *   %esi : source physical address
 *   %edi : destination physical address
 *   %ecx : length
 *   %bx : memcpy()-like function to call, passing parameters:
 *	     %ds:esi : source address
 *	     %es:edi : destination address
 *	     %ecx : length
 *         and returning:
 *	     %ds:esi : next source address
 *	     %es:edi : next destination address
 * Returns:
 *   %esi : next source physical address
 *   %edi : next destination physical address
 * Corrupts:
 *   None
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
process_bytes:

#ifndef KEEP_IT_REAL

	/* Preserve registers */
	pushfw
	pushl	%eax
	pushl	%ebp

	/* Construct GDT on stack (since .prefix may not be writable) */
	.equ	PM_DS, 0x18	/* Flat data segment */
	pushl	$0x008f9300
	pushl	$0x0000ffff
	.equ	PM_SS, 0x10	/* Stack segment based at %ss:0000 */
	pushl	$0x008f0930
	pushw	%ss
	pushw	$0xffff
	.equ	PM_CS, 0x08	/* Code segment based at %cs:0000 */
	pushl	$0x008f09b0
	pushw	%cs
	pushw	$0xffff
	pushl	$0		/* Base and length */
	pushw	%ss
	pushw	$0x1f
	movzwl	%sp, %ebp
	shll	$4, 0x02(%bp)
	addl	%ebp, 0x02(%bp)
	shll	$4, 0x0a(%bp)
	shll	$4, 0x12(%bp)
	subw	$8, %sp
	sgdt	-8(%bp)

	/* Switch to protected mode */
	pushw	%gs
	pushw	%fs
	pushw	%es
	pushw	%ds
	pushw	%ss
	pushw	%cs
	pushw	$2f
	cli
	data32 lgdt (%bp)
	movl	%cr0, %eax
	orb	$CR0_PE, %al
	movl	%eax, %cr0
	ljmp	$PM_CS, $1f
1:	movw	$PM_SS, %ax
	movw	%ax, %ss
	movw	$PM_DS, %ax
	movw	%ax, %ds
	movw	%ax, %es
	movw	%ax, %fs
	movw	%ax, %gs

	/* Call memcpy()-like function */
	call	*%bx

	/* Return to (flat) real mode */
	movl	%cr0, %eax
	andb	$0!CR0_PE, %al
	movl	%eax, %cr0
	lret
2:	/* lret will ljmp to here */
	popw	%ss
	popw	%ds
	popw	%es
	popw	%fs
	popw	%gs

	/* Restore GDT */
	data32 lgdt -8(%bp)
	addw	$( 8 /* saved GDT */ + ( PM_DS + 8 ) /* GDT on stack */ ), %sp

	/* Restore registers and return */
	popl	%ebp
	popl	%eax
	popfw
	ret

#else /* KEEP_IT_REAL */

	/* Preserve registers */
	pushl	%eax
	pushw	%ds
	pushw	%es
	
	/* Convert %esi and %edi to %ds:esi and %es:edi */
	shrl	$4, %esi
	movw	%si, %ds
	xorw	%si, %si
	shll	$4, %esi
	shrl	$4, %edi
	movw	%di, %es
	xorw	%di, %di
	shll	$4, %edi

	/* Call memcpy()-like function */
	call	*%bx

	/* Convert %ds:esi and %es:edi back to physical addresses */
	xorl	%eax, %eax
	movw	%ds, %cx
	shll	$4, %eax
	addl	%eax, %esi
	xorl	%eax, %eax
	movw	%es, %cx
	shll	$4, %eax
	addl	%eax, %edi

	/* Restore registers and return */
	popw	%es
	popw	%ds
	popl	%eax
	ret

#endif /* KEEP_IT_REAL */

	.size	process_bytes, . - process_bytes

/****************************************************************************
 * install_block
 *
 * Install block to specified address
 *
 * Parameters:
 *   %esi : source physical address (must be a multiple of 16)
 *   %edi : destination physical address (must be a multiple of 16)
 *   %ecx : length of (decompressed) data
 *   %edx : total length of block (including any uninitialised data portion)
 * Returns:
 *   %esi : next source physical address (will be a multiple of 16)
 *   %edi : next destination physical address (will be a multiple of 16)
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
install_block:
	/* Preserve registers */
	pushl	%ecx
	pushw	%bx

	/* Decompress (or copy) source to destination */
#if COMPRESS
	movw	$decompress16, %bx
#else
	movw	$copy_bytes, %bx
#endif
	call	process_bytes

	/* Zero .bss portion */
	negl	%ecx
	addl	%edx, %ecx
	movw	$zero_bytes, %bx
	call	process_bytes

	/* Round up %esi and %edi to start of next blocks */
	addl	$0xf, %esi
	andl	$~0xf, %esi
	addl	$0xf, %edi
	andl	$~0xf, %edi

	/* Restore registers and return */
	popw	%bx
	popl	%ecx
	ret
	.size install_block, . - install_block

	.section ".prefix.lib", "awx", @progbits

#ifdef EBDA_DEBUG

text16_str:
	.asciz	"TEXT16="
data16_str:
	.asciz	" DATA16="
FBMS_str:
	.asciz	"FBMS="
EBDA_str:
	.asciz	" EBDA="
int15_c1h_str:
	.asciz	" INT15/C1h="

/****************************************************************************
 * Parameters: 
 *   %ax : .text16 segment address
 *   %bx : .data16 segment address
 * Corrupts:
 *   none
 ****************************************************************************
 */
print_section_state:
	pushw	%ax
	pushw	%di
	pushw	%si
	push	%ds
	pushw	%cs
	popw	%ds
	xorw	%di, %di
	movw	$text16_str, %si
	call	print_message
	call	print_hex_word
	movw	$data16_str, %si
	call	print_message
	movw	%bx, %ax
	call	print_hex_word
	movb	$' ', %al
	call	print_character
	popw	%ds
	popw	%si
	popw	%di
	popw	%ax
	/* Fall-thru */
print_ebda_state:	
	pushw	%ax
	pushw	%di
	pushw	%si
	push	%ds
	pushw	%fs
	pushw	%cs
	popw	%ds
	xorw	%di, %di
	pushw	$0x40
	popw	%fs
	movw	$FBMS_str, %si
	call	print_message
	movw	%fs:0x13, %ax
	shlw	$6, %ax
	call	print_hex_word
	movw	$EBDA_str, %si
	call	print_message
	movw	%fs:0x0e, %ax
	call	print_hex_word
	movb	$',', %al
	call	print_character
	pushw	%fs:0x0e
	popw	%fs
	movb	%fs:0, %al
	call	print_hex_byte
	movw	$int15_c1h_str, %si
	call	print_message
	movb	$0xc1, %ah
	pushw	%es
	int	$0x15
	movw	%es, %ax
	popw	%es
	call	print_hex_word
	movb	$0x0a, %al
	call	print_character
	popw	%fs
	popw	%ds
	popw	%si
	popw	%di
	popw	%ax
	ret
#endif /* EBDA_DEBUG */

#ifdef EBDA_HEXDUMP

/****************************************************************************
 * Hexdump the first 4 non-empty paragraphs, and the last 4 non-empty
 * paragraphs of the given region
 *
 * Parameters: 
 *   %ax : start segment
 *   %bx : end segment
 * Corrupts:
 *   none
 ****************************************************************************
 */
hexdump:
	pushaw
	pushw	%es
	movw	%ax, %bp
	xorw	%di, %di
	xorw	%dx, %dx

1:	/* %bp => current segment */
	cmp	%bp, %bx
	je	99f
	movw	%bp, %es
	/* Check if paragraph is non-zero */
	xorw	%si, %si
	movw	$0x10, %cx
2:	cmpb	$0, %es:0(%si)
	jne	3f
	inc	%si
	loop	2b
	/* Skip this line */
	inc	%bp
	jmp	1b
3:	/* Print out segment */
	movw	%bp, %ax
	call	print_hex_word
	movb	$':', %al
	call	print_character
	movb	$' ', %al
	call	print_character
	/* Print out data */
	xorw	%si, %si
	movw	$0x10, %cx
4:	movb	%es:0(%si), %al
	call	print_hex_byte
	movb	$' ', %al
	call	print_character
	inc	%si
	loop	4b
	/* Move onto the next paragraph */
	inc	%bp
	movb	$0xa, %al
	call	print_character
	/* Have we finished the bottom? */
	inc	%dx
	cmp	$4, %dx
	jne	1b

	/* Find the last four non-zero paragraphs, %ax => search
	 * point, starting at %bx down to %bp */
10:	xor	%dx, %dx
	movw	%bx, %ax
	dec	%ax
11:	cmp	%ax, %bp
	je	20f
	movw	%ax, %es
	/* Check if paragraph is non-zero */
	xorw	%si, %si
	movw	$0x10, %cx
12:	cmpb	$0, %es:0(%si)
	jne	13f
	inc	%si
	loop	12b
	/* Skip this line */
	dec	%ax
	jmp	11b
13:	/* Is this our fourth non-zero final paragraph? */
	dec	%ax
	inc	%dx
	cmp	$4, %dx
	jne	11b

20:	/* Recurse. %ax and %bx are conveniently in the correct format */
	call	hexdump

99:	popw	%es
	popaw
	ret
	
#endif	/* EBDA_HEXDUMP */

/****************************************************************************
 * alloc_basemem
 *
 * Allocate space for .text16 and .data16 from top of base memory.
 * Memory is allocated using the BIOS free base memory counter at
 * 0x40:13.
 *
 * Parameters: 
 *   none
 * Returns:
 *   %ax : .text16 segment address
 *   %bx : .data16 segment address
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	alloc_basemem
alloc_basemem:
	/* Preserve registers */
	pushw	%fs
#if defined(EBDA_ALLOCATE) || defined(EBDA_RELOCATE)
	pushw	%ds
	pushw	%es
	pushw	%cx
	pushw	%dx
	pushw	%si
	pushw	%di

	pushw	$0x40
	popw	%fs

	/* Only use EBDA when iSCSI booting, ie. when called from bcv_entry().
	 * %ecx : exec() entry point */
	cmpl	$instance_bcv_main, %ecx
	jne	use_fbms

#ifdef EBDA_DEBUG
	call	print_ebda_state
#endif
	/* (%bx=%es):0 => Old EBDA base */
	movw	%fs:0x0e, %bx
	movw	%bx, %es
	/* Determine if EBDA is present on this platform */
	cmp	$0, %bx
	je	use_fbms
	/* %dx => old EBDA size/kbytes */
	xorw	%dx, %dx
	movb	%es:0, %dl

#ifdef EBDA_HEXDUMP
	/* Print out the existing EBDA */
	pushw	%ax
	pushw	%bx
	movw	%bx, %ax
	movw	%dx, %bx
	shlw	$6, %bx
	addw	%ax, %bx
	call	hexdump
	popw	%bx
	popw	%ax
#endif	/* EBDA_HEXDUMP */

#if defined(EBDA_ALLOCATE)
	/* Check EBDA has room */
	movw	%dx, %ax
	addw	$_textdata16_memsz_kbyte, %ax
	cmpw	$MAX_EBDA_KBYTES, %ax
	ja	use_fbms

	/* Store the new EBDA size */
	movb	%al, %es:0
#endif	/* EBDA_ALLOCATE */

	/* %ax => new base of EBDA */
	movw	%bx, %ax
	subw	$_textdata16_memsz_pgh, %ax
	movw	%ax, %fs:0x0e

	/* Copy %dx kbytes from %bx:0 down to %ax:0. For safety's sake, don't
	 * assume that EBDA fits into a single segment */
	pushw	%dx
1:	cmpw	$0, %dx
	je	2f
	movw	%bx, %ds
	xorw	%si, %si
	movw	%ax, %es
	xorw	%di, %di
	movw	$0x400, %cx
	rep movsb
	addw	$0x40, %ax
	addw	$0x40, %bx
	dec	%dx
	jmp	1b

2:	popw	%dx
	/* .text16 sits above the newly moved EBDA */
	movw	%fs:0x0e, %ax
	shlw	$6, %dx
	addw	%dx, %ax
	jmp	99f

use_fbms:
	/* .text16 sits at the bottom of the newly allocated base memory */
	movw	%fs:0x13, %ax
	subw	$_textdata16_memsz_kbyte, %ax
	shlw	$6, %ax

99:	/* Allocate free base memory */
	movw	%fs:0x13, %cx
	subw	$_textdata16_memsz_kbyte, %cx
	movw	%cx, %fs:0x13

	/* .data16 sits just above .text16 */
	movw	%ax, %bx
	addw	$_text16_memsz_pgh, %bx

#ifdef EBDA_HEXDUMP
	/* Print out the moved old region of EBDA */
	pushw	%ax
	pushw	%bx
	movw	%fs:0x0e, %ax
	movw	%ax, %bx
	addw	%dx, %bx	/* %dx is already in kbytes */
	call	hexdump
	popw	%bx
	popw	%ax
#endif	/* EBDA_HEXDUMP */

	/* Restore registers and return */
	popw	%di
	popw	%si
	popw	%dx
	popw	%cx
	popw	%es
	popw	%ds
#ifdef EBDA_DEBUG
	call	print_section_state
#endif

#else	/* EBDA_ALLOCATE || EBDA_RELOCATE */

	/* FBMS => %ax as segment address */
	pushw	$0x40
	popw	%fs
	movw	%fs:0x13, %ax
	shlw	$6, %ax

	/* Calculate .data16 segment address */
	subw	$_data16_memsz_pgh, %ax
	pushw	%ax

	/* Calculate .text16 segment address */
	subw	$_text16_memsz_pgh, %ax
	pushw	%ax

	/* Update FBMS */
	shrw	$6, %ax
	movw	%ax, %fs:0x13

	/* Retrieve .text16 and .data16 segment addresses */
	popw	%ax
	popw	%bx

	/* Restore registers and return */
	
#endif	/* EBDA_ALLOCATE || EBDA_RELOCATE */

	popw	%fs
	ret
	.size alloc_basemem, . - alloc_basemem

/****************************************************************************
 * free_basemem
 *
 * Free space allocated with alloc_basemem.
 *
 * Parameters:
 *   %ax : .text16 segment address
 *   %bx : .data16 segment address
 * Returns:
 *   %ax : 0 if successfully freed
 * Corrupts:
 *   none
 ****************************************************************************
 */
#if defined(EBDA_ALLOCATE) || defined(EBDA_RELOCATE)
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	free_basemem
free_basemem:
	/* Preserve registers */
	pushw	%fs
	pushw	%gs
	pushw	%ds
	pushw	%es
	pushw	%bx
	pushw	%cx
	pushw	%dx

	pushw	$0x40
	popw	%fs
	/* %gs => .text16 */
	movw	%ax, %gs
#ifdef EBDA_DEBUG
	call	print_section_state
#endif

	/* Check hooked interrupt count */
	cmpw	$0, %gs:hooked_bios_interrupts
	jne	99f

	/* (%bx=%ds):0 => bottom of EBDA */
	movw	%fs:0x0e, %bx
	movw	%bx, %ds
	/* Check if EBDA is present */
	cmpw	$0, %bx
	je	check_fbms
	/* Check that .text16 is above the base of EBDA. If it isn't, then we
	 * used FBMS instead of EBDA for our allocation. */
	cmpw	%ax, %bx
	ja	check_fbms
	/* Check that we're still the final EBDA allocation. If there's been
	 * any additional allocation then it's unsafe to free up *any* memory */
	xorw	%dx, %dx
	movb	%ds:0, %dl
	/* %dx => Remaining EBDA size_kbytes */
#ifdef EBDA_ALLOCATE
	subw	$_textdata16_memsz_kbyte, %dx
#endif
	movw	%dx, %cx
	shlw	$6, %cx
	addw	%bx, %cx
	cmp	%cx, %ax
	jne	99f

	/* Check that nobody has allocated FBMS below EBDA */
	movw	%fs:0x13, %cx
	shlw	$6, %cx
	cmpw	%cx, %bx
	jne	99f

#ifdef EBDA_ALLOCATE
	/* Reduce the EBDA size, and move the base of EBDA up */
	movb	%dl, %ds:0
#endif
	movw	%bx, %ax
	addw	$_textdata16_memsz_pgh, %bx
	movw	%bx, %fs:0x0e
	/* Move %dx kbytes of the remaining EBDA region from %ax to %bx.
	 * Since these regions might overlap, do the copy top-down */
	shlw	$6, %dx
	addw	%dx, %ax
	addw	%dx, %bx
	shrw	$6, %dx
1:	cmpw	$0, %dx
	je	free_fbms
	subw	$0x40, %ax
	subw	$0x40, %bx
	dec	%dx
	movw	%ax, %ds
	xorw	%si, %si
	movw	%bx, %es
	xorw	%di, %di
	movw	$0x400, %cx
	rep movsb
	jmp	1b

check_fbms:
	/* Check that nobody else has allocated FBM below us */
	movw	%fs:0x13, %cx
	shlw	$6, %cx
	cmpw	%cx, %ax
	jne	99f

free_fbms:
	/* Free base memory */
	movw	%fs:0x13, %ax
	addw	$_textdata16_memsz_kbyte, %ax
	movw	%ax, %fs:0x13

	xorw	%ax, %ax

#ifdef EBDA_HEXDUMP
	/* hexdump all of EBDA */
	pushw	%fs
	pushw	%ax
	pushw	%bx
	pushw	$0x40
	popw	%fs
	movw	%fs:0x0e, %ax
	movw	%ax, %fs
	movw	%fs:0, %bx
	shlw	$6, %bx
	addw	%ax, %bx
	call	hexdump
	popw	%bx
	popw	%ax
	popw	%fs
#endif	

99:	/* Restore registers and return */
	popw	%dx
	popw	%cx
	popw	%bx
	popw	%es
	popw	%ds
	popw	%gs
	popw	%fs
#ifdef EBDA_DEBUG
	call	print_ebda_state
#endif
	ret
	.size free_basemem, . - free_basemem

#else	/* EBDA_ALLOCATE || EBDA_RELOCATE */

	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl	free_basemem
free_basemem:
	/* Preserve registers */
	pushw	%fs
	pushw	%gs

	/* Check FBMS counter */
	pushw	%ax
	shrw	$6, %ax
	pushw	$0x40
	popw	%fs
	cmpw	%ax, %fs:0x13
	popw	%ax
	jne	1f

	/* Check hooked interrupt count */
	movw	%ax, %gs
	cmpw	$0, %gs:hooked_bios_interrupts
	jne	1f

	/* OK to free memory */
	addw	$_text16_memsz_pgh, %ax
	addw	$_data16_memsz_pgh, %ax
	shrw	$6, %ax
	movw	%ax, %fs:0x13

	xorw	%ax, %ax

1:	/* Restore registers and return */
	popw	%gs
	popw	%fs
	ret
	.size free_basemem, . - free_basemem
#endif	/* EBDA_ALLOCATE || EBDA_RELOCATE */

	.section ".text16.data", "aw", @progbits
	.globl	hooked_bios_interrupts
hooked_bios_interrupts:
	.word	0
	.size	hooked_bios_interrupts, . - hooked_bios_interrupts

#ifndef SFC_SPACE_SAVING_HACKS
/****************************************************************************
 * install
 *
 * Install all text and data segments.
 *
 * Parameters:
 *   none
 * Returns:
 *   %ax  : .text16 segment address
 *   %bx  : .data16 segment address
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl install
install:
	/* Preserve registers */
	pushl	%esi
	pushl	%edi
	pushl	%ebp

	/* Allocate space for .text16 and .data16 */
	call	alloc_basemem
	/* Image source = %cs:0000 */
	xorl	%esi, %esi

	/* Image destination = default */
	xorl	%edi, %edi
	/* Allow arbitrary relocation */
	orl	$0xffffffff, %ebp
	/* Install text and data segments */
	call	install_prealloc
	/* Restore registers and return */
	popl	%ebp
	popl	%edi
	popl	%esi
	ret
	.size install, . - install
#endif /* SFC_SPACE_SAVING_HACKS */

/****************************************************************************
 * install_prealloc
 *
 * Install all text and data segments.
 *
 * Parameters:
 *   %ax  : .text16 segment address
 *   %bx  : .data16 segment address
 *   %esi : Image source physical address (or zero for %cs:0000)
 *   %edi : Decompression temporary area physical address (or zero for default)
 *   %ebp : Maximum end address for relocation
 *          - 0xffffffff for no maximum
 *          - 0x00000000 to inhibit use of INT 15,e820 and INT 15,e801
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.code16
	.globl install_prealloc
install_prealloc:
	/* Save registers on external stack */
	pushal
	pushw	%ds
	pushw	%es
	cld			/* Sanity: clear the direction flag asap */

	/* Switch to temporary stack in .bss16 */
	pushw	%ss
	popw	%ds
	movl	%esp, %ecx
	movw	%bx, %ss
	movl	$_data16_memsz, %esp
	pushw	%ds
	pushl	%ecx

	/* Set up %ds for (read-only) access to .prefix */
	pushw	%cs
	popw	%ds

	/* Save decompression temporary area physical address */
	pushl	%edi

	/* Install .text16.early and calculate %ecx as offset to next block */
	pushl	%esi
	xorl	%esi, %esi
	movw	%cs, %si
	shll	$4, %esi
	pushl	%esi		/* Save original %cs:0000 */
	addl	$_text16_early_lma, %esi
	movzwl	%ax, %edi
	shll	$4, %edi
	movl	$_text16_early_filesz, %ecx
	movl	$_text16_early_memsz, %edx
	call	install_block	/* .text16.early */
	popl	%ecx		/* Calculate offset to next block */
	subl	%esi, %ecx
	negl	%ecx
	popl	%esi

#ifndef KEEP_IT_REAL

	/* Access high memory by enabling the A20 gate.  (We will
	 * already have 4GB segment limits as a result of calling
	 * install_block.)
	 */
	pushw	%cs
	pushw	$1f
	pushw	%ax
	pushw	$access_highmem
	lret
1:	/* Die if we could not access high memory */
	jnc	3f
	movw	$a20_death_message, %si
	xorw	%di, %di
	call	print_message
2:	jmp	2b
	.section ".prefix.data", "aw", @progbits
a20_death_message:
	.asciz	"\nHigh memory inaccessible - cannot continue\n"
	.size	a20_death_message, . - a20_death_message
	.previous
3:
	/* Restore decompression temporary area physical address.  Store the
	 * next destination physical address (%edi) to be restored later */
	movl	%edi, %ecx
	popl	%edi

	/* Find a suitable decompression temporary area, if none specified */
	pushl	%eax
	testl	%edi, %edi
	jnz	1f
	/* Use INT 15,88 to find the highest available address via INT
	 * 15,88.  This limits us to around 64MB, which should avoid
	 * all of the POST-time memory map failure modes.
	 */
	movb	$0x88, %ah
	int	$0x15
	movw	%ax, %di
	addl	$0x400, %edi
	subl	$_textdata_memsz_kb, %edi
	shll	$10, %edi
	/* Sanity check: if we have ended up below 1MB, use 1MB */
1:	cmpl	$0x100000, %edi
	jae	1f
	movl	$0x100000, %edi
1:	testl	%ebx, %ebx
	jnz	2f
2:	popl	%eax
#endif
	/* Save decompression temporary area physical address and restore
	 * the next destination physical address to %edi */
	pushl	%edi
	movl	%ecx, %edi

	/* Open payload (which may not yet be in memory) */
	pushw	%cs
	pushw	$1f
	pushw	%ax
	pushw	$open_payload
	lret
1:	/* Die if we could not access the payload */
	jnc	3f
	xorw	%di, %di
	movl	%esi, %eax
	call	print_hex_dword
	movw	$payload_death_message, %si
	call	print_message
2:	jmp	2b
	.section ".prefix.data", "aw", @progbits
payload_death_message:
	.asciz	"\nPayload inaccessible - cannot continue\n"
	.size	payload_death_message, . - payload_death_message
	.previous
3:

	/* Calculate physical address of payload (i.e. first source) */
	testl	%esi, %esi
	jnz	1f
	movw	%cs, %si
	shll	$4, %esi
1:	addl	%ecx, %esi

	/* Install .text16.late and .data16 */
	movl	$_text16_late_filesz, %ecx
	movl	$_text16_late_memsz, %edx
	call	install_block	/* .text16.late */
	movzwl	%bx, %edi
	shll	$4, %edi
	movl	$_data16_filesz, %ecx
	movl	$_data16_filesz, %edx	/* do not zero our temporary stack */
	call	install_block		/* .data16 */

	/* Set up %ds for access to .data16 */
	movw	%bx, %ds

	/* Restore decompression temporary area physical address */
	popl	%edi

#ifndef KEEP_IT_REAL

	/* Install .text and .data to temporary area in high memory,
	 * prior to reading the E820 memory map and relocating
	 * properly.
	 */
	pushl	%edi
	movl	$_textdata_filesz, %ecx
	movl	$_textdata_memsz, %edx
	call	install_block
	popl	%edi

#endif /* KEEP_IT_REAL */

	/* Switch back to original stack and zero .bss16 */
	lss	%ss:(%esp), %esp
	pushl	%edi
	pushw	%es
	movw	%bx, %es
	movl	$_data16_filesz, %edi
	movl	$_data16_memsz, %ecx
	subl	%edi, %ecx
	call	zero_bytes
	popw	%es
	popl	%edi

#ifndef KEEP_IT_REAL

	/* Initialise librm at current location */
	movw	%ax, (init_librm_vector+2)
	lcall	*init_librm_vector

	/* Inhibit INT 15,e820 and INT 15,e801 if applicable */
	testl	%ebp, %ebp
	jnz	1f
	incb	memmap_post
	decl	%ebp
1:
	/* Call relocate() to determine target address for relocation.
	 * relocate() will return with %esi, %edi and %ecx set up
	 * ready for the copy to the new location.
	 */
	movw	%ax, (prot_call_vector+2)
	pushl	$relocate
	lcall	*prot_call_vector
	popl	%edx /* discard */

	/* Copy code to new location */
	pushl	%edi
	pushw	%bx
	movw	$copy_bytes, %bx
	call	process_bytes
	popw	%bx
	popl	%edi

	/* Initialise librm at new location */
	lcall	*init_librm_vector

#else /* KEEP_IT_REAL */

	/* Initialise libkir */
	movw	%ax, (init_libkir_vector+2)
	lcall	*init_libkir_vector

#endif /* KEEP_IT_REAL */

	/* Close access to payload */
	movw	%ax, (close_payload_vector+2)
	lcall	*close_payload_vector

	/* Restore registers */
	popw	%es
	popw	%ds
	popal
	ret
	.size install_prealloc, . - install_prealloc

	/* Vectors for far calls to .text16 functions.  Must be in
	 * .data16, since .prefix may not be writable.
	 */
	.section ".data16", "aw", @progbits
#ifdef KEEP_IT_REAL
init_libkir_vector:
	.word init_libkir
	.word 0
	.size init_libkir_vector, . - init_libkir_vector
#else
init_librm_vector:
	.word init_librm
	.word 0
	.size init_librm_vector, . - init_librm_vector
prot_call_vector:
	.word prot_call
	.word 0
	.size prot_call_vector, . - prot_call_vector
#endif
close_payload_vector:
	.word close_payload
	.word 0
	.size close_payload_vector, . - close_payload_vector

	/* Dummy routines to open and close payload */
	.section ".text16.early.data", "aw", @progbits
	.weak	open_payload
	.weak	close_payload
open_payload:
close_payload:
	clc
	lret
	.size	open_payload, . - open_payload
	.size	close_payload, . - close_payload

/****************************************************************************
 * uninstall
 *
 * Uninstall all text and data segments.
 *
 * Parameters:
 *   %ax  : .text16 segment address
 *   %bx  : .data16 segment address
 * Returns:
 *   none
 * Corrupts:
 *   none
 ****************************************************************************
 */
#if defined(EBDA_ALLOCATE) || defined(EBDA_RELOCATE)
	.section ".prefix.lib", "awx", @progbits
#else
	.section ".text16", "ax", @progbits
#endif
	.code16
	.globl uninstall
uninstall:
	call	free_basemem
	ret
	.size uninstall, . - uninstall



	/* File split information for the compressor */
#if COMPRESS
#define PACK_OR_COPY "PACK"
#else /* COMPRESS */
#define PACK_OR_COPY "COPY"
#endif /* COMPRESS */
	.section ".zinfo", "a", @progbits
	.ascii	"COPY"
	.long	_prefix_lma
	.long	_prefix_filesz
	.long	_max_align
	.ascii	PACK_OR_COPY
	.long	_text16_early_lma
	.long	_text16_early_filesz
	.long	_max_align
	.ascii	"PAYL"
	.long	0
	.long	0
	.long	_payload_align
	.ascii	"COPY"
	.long	_pprefix_lma
	.long	_pprefix_filesz
	.long	_max_align
	.ascii	PACK_OR_COPY
	.long	_text16_late_lma
	.long	_text16_late_filesz
	.long	_max_align
	.ascii	PACK_OR_COPY
	.long	_data16_lma
	.long	_data16_filesz
	.long	_max_align
	.ascii	PACK_OR_COPY
	.long	_textdata_lma
	.long	_textdata_filesz
	.long	_max_align

	.weak	_payload_align
	.equ	_payload_align, 1

#if defined(EBDA_ALLOCATE) || defined(EBDA_RELOCATE)

#define STACK_SIZE_KBYTE 0x1

/****************************************************************************
 * Allocate stack16 (near mode real-call)
 *
 * EBDA is a precious resource, offically limited to 64k. Our stack16 does
 * not need to go be inside EBDA, because it's unused for any interrupts
 *
 * So instead, place it at the top of FBMS. Modify 40:13h so we don't overlap
 * with gPXE buffers allocated in the same fashion. This reduces textdata16
 * from 13k to 9k
 *
 * Parameters:
 *   none
 * Returns:
 *   %si: Stack segment
 *   %di: Stack offset
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits
	.global allocate_stack
allocate_stack:
	pushw	%fs
	pushw	$0x40
	popw	%fs

	movw	$STACK_SIZE_KBYTE, %di
	movw	%fs:0x13, %si
	subw	%di, %si
	movw	%si, %fs:0x13

	shlw	$6, %si
	shlw	$10, %di
	popw	%fs
	ret

/****************************************************************************
 * Free stack16 (near mode real-call)
 *
 * Parameters:
 *   %ax: text16 segment
 *   %si: Stack segment
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.global	release_stack
release_stack:
	pushw	%fs
	pushw	%si
	pushw	$0x40
	popw	%fs

	/* Check that there's been no other allocations */
	shrw	$6, %si
	cmp	%fs:0x13, %si
	jne	99f
	movw	%fs:0x13, %si
	/* Free FBMS */
	addw	$STACK_SIZE_KBYTE, %si
	movw	%si, %fs:0x13

99:	popw	%si
	popw	%fs
	ret

#endif	/* EBDA_ALLOCATE || EBDA_RELOCATE */

	.section ".text16.data", "aw", @progbits
	.global	prefix_seg
prefix_seg:
	.word	0x0000

#ifdef EBDA_ALLOCATE

#define EBDA_SIGNATURE 0x62416645

/****************************************************************************
 * text16 signature and pci_busdevfn
 *
 * To support EBDA, each interrupt handler has a stub inside the prefix.
 * Unfortunately, there is no place suitable for storing the offset from the
 * base of EBDA to our text16 segment. We can't rely on PMM, and the prefix
 * is read-only from the BCV/BEV.
 *
 * So, we have to search for it. text16 is identified by a run-time signature,
 * and by the code segment of the (calling) prefix
 *
 ****************************************************************************
 */
text16_signature:
	.long	EBDA_SIGNATURE
	.size	text16_signature, . - text16_signature

/****************************************************************************
 * find text16 segment (real-mode near call)
 *
 * Search EBDA for the text16/data16 segment
 *
 * Parameters:
 *   %cs  : Code segment of matching prefix
 * Returns:
 *   %ax  : .text16 segment address
 *   %bx  : .data16 segment address
 * Corrupts:
 *   none
 ****************************************************************************
 */
	.section ".prefix.lib", "awx", @progbits

	.globl	find_text16
find_text16:
	pushw	%gs
	pushw	%fs
	pushw	%dx
	pushw	$0x40
	popw	%fs

	/* The .text16 segment could be in EBDA or between the bottom of EBDA
	 * and the top of FBMS.  Search downwards for the signature from (%ax=)
	 * the top of EBDA to (%dx=) FBMS.
	 *
	 * Consider EBDA to be corrupt if any of the following are true:
	 *  - the base of EBDA is zero
	 *  - the base of EBDA is lower than FBMS
	 *  - both EBDA and the "gap" between EBDA and FBMS are too small for
	 *    the .text16 and .data16 blocks */
	movw	%fs:0x13, %bx
	shlw	$6, %bx
	movw	%fs:0x0e, %dx
	cmpw	$0, %dx
	je	ebda_corrupt
	/* (%dx=%gs):0 => bottom of EBDA */
	movw	%dx, %gs
	xorw	%ax, %ax
	movb	%gs:0, %al
	cmpb	$_textdata16_memsz_kbyte, %al
	jae	1f		/* EBDA is big enough */
	subw	%bx, %dx
	jc	ebda_corrupt	/* FBMS > EBDA */
	cmpw	$_textdata16_memsz_pgh, %dx
	jb	ebda_corrupt	/* FBMS too small */
1:	/* EBDA looks valid. Find the top of it */
	pushw	%gs
	popw	%dx
	shlw	$6, %ax
	addw	%dx, %ax
	movw	%bx, %dx
	jmp	search

ebda_corrupt:
	/* Search down from (%ax=) 0xa000:0 down to (%dx=) the minimum of
	 * FBMS, the base of EBDA and 0xa000 - MAX_EBDA_KBYTES */
	movw	%fs:0x13, %dx
	shlw	$6, %dx
	/* Some bioses set %ebda:0 = 0 whilst preserving the contents, so
	 * so only validate the address */
	movw	%fs:0x0e, %ax
	cmpw	$0, %ax
	je	1f
	cmpw	%ax, %dx
	jb	1f
	movw	%ax, %dx
1:	movw	$(0xa000 - MAX_EBDA_KBYTES), %ax
	cmpw	%ax, %dx
	jb	2f
	movw	%ax, %dx
2:	movw	$0xa000, %ax

search:
	subw	$_textdata16_memsz_pgh, %ax
	/* Search for prefix %cs => %bx */
	movw	%cs, %bx

	dec	%dx
	inc	%ax
1:	dec	%ax
	cmpw	%dx, %ax
	jb	98f
	movw	%ax, %gs
	/* Check signature */
	cmpl	$EBDA_SIGNATURE, %gs:text16_signature
	jne	1b
	/* Check prefix code segment */
	cmpw	%bx, %gs:prefix_seg
	jne	1b
	/* Found matching section. %ax = text16 */
	movw	%ax, %bx
	addw	$_text16_memsz_pgh, %bx
	jmp	99f

98:	/* Failed. Zero out %ax, %bx */
	xorw	%ax, %ax
	xorw	%bx, %bx
99:	popw	%dx
	popw	%fs
	popw	%gs
	ret
	.size find_text16, . - find_text16

/****************************************************************************
 * Find the text16 segment and jump to the given offset (real mode near-call)
 *
 * Called from interrupt handlers. Searches the EBDA region for text16/data16
 * segment, and, it if finds it, jump to the offset pushed on the stack. If
 * we can't find the text16 segment, then print an error message and hang.
 * Hell, that's more than you get when the interrupt handler has moved
 * normally..
 *
 * Parameters:
 *   none
 * Stack:
 *    4(%sp) Current %cs
 *    2(%sp) Offset of bottom half
 *    0(%sp) Offset of top half
 * Returns:
 *   none
 * Corrupts:
 *   none
 ****************************************************************************
 */

lost_message:
	.asciz	"Oops! Unable to find realmode segment\n"
	.size lost_message, . - lost_message

	.globl	find_and_jump_text16
find_and_jump_text16:
	pushw	%ax
	pushw	%bx
	pushw	%ds
	pushl	%edi

	call	find_text16
	cmp	%ax, %bx
	je	98f

	/* Found it. Jump to the text16 segment */
	pushw	%ax
	pushw	$1f
	lret
	
	.section ".text16", "ax", @progbits
1:	/* Initialse librm using the callers stack */
	movw	%bx, %ds
	movl	_virt_offset, %edi
	pushw	%cs
	call	init_librm

	popl	%edi
	popw	%ds
	popw	%bx

	pushw	%bp
	movw	%sp, %bp

	/* The stack now goes:
	 *   0(%sp)     old %bp
	 *   2(%sp)	old %ax
	 *   4(%sp)     return offset
	 *   6(%sp)	offset of bottom half
	 *   8(%sp)     segment of top half
	 *  10(%sp)     flags
	 */
	movw	%ax, 8(%bp)
	popw	%bp
	popw	%ax

	/* Stack now goes:
	 *  %2(sp)     offset of bottom half
	 *  %4(sp)     segment of bottom half
	 *  %6(sp)     flags
	 * so once we add 2 to the stack pointer, the stack is left in the
	 * correct state for an "iret" instruction to jump to the bottom
	 * half with the flags restored.
	 */
	add	$2, %sp
	iret

	.previous
98:	/* Failed finding text16 segment */
	pushw	%cs
	popw	%ds
#ifdef EBDA_DEBUG
	call	print_ebda_state
#endif
	xor	%di, %di
	movw	$lost_message, %si
	call	print_message
99:	cli
	/* Spin forever */
	jmp 99b

#endif	/* EBDA_ALLOCATE */
