/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef CONFIG_GENERAL_H
#define CONFIG_GENERAL_H

/** @file
 *
 * General configuration
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <config/defaults.h>

/*
 * Branding
 *
 * Vendors may use these strings to add their own branding to gPXE.
 * PRODUCT_NAME is displayed prior to any gPXE branding in startup
 * messages, and PRODUCT_SHORT_NAME is used where a brief product
 * label is required (e.g. in BIOS boot selection menus).
 *
 * To minimise end-user confusion, it's probably a good idea to either
 * make PRODUCT_SHORT_NAME a substring of PRODUCT_NAME or leave it as
 * "gPXE".
 *
 */
#define PRODUCT_VERSION "v5.2.2.1004"
#define PRODUCT_NAME "Solarflare Boot Manager (" PRODUCT_VERSION ")"
#define PRODUCT_AUTHOR "Solarflare Communications 2008-2019"
#define PRODUCT_SHORT_NAME "Solarflare"

/*
 * EFI platform is tightly controlled. Don't need much gPXE functionality
 *
 */
#if ! defined(EFI_BUILD)

/*
 * Timer configuration
 *
 */
#define BANNER_TIMEOUT	20	/* Tenths of a second for which the shell
				   banner should appear */

/*
 * Network protocols
 *
 */

#define	NET_PROTO_IPV4		/* IPv4 protocol */
#define	NET_PROTO_ICMPV4	/* ICMPv4 protocol */

/*
 * Download protocols
 *
 */

#define	DOWNLOAD_PROTO_TFTP	/* Trivial File Transfer Protocol */
#undef	DOWNLOAD_PROTO_NFS	/* Network File System */
#define	DOWNLOAD_PROTO_HTTP	/* Hypertext Transfer Protocol */
#undef	DOWNLOAD_PROTO_HTTPS	/* Secure Hypertext Transfer Protocol */
#undef	DOWNLOAD_PROTO_FTP	/* File Transfer Protocol */
#undef	DOWNLOAD_PROTO_TFTM	/* Multicast Trivial File Transfer Protocol */
#undef	DOWNLOAD_PROTO_SLAM	/* Scalable Local Area Multicast */
#undef	DOWNLOAD_PROTO_FSP	/* FSP? */

/*
 * SAN boot protocols
 *
 */

#undef	SANBOOT_PROTO_ISCSI	/* iSCSI protocol */
#undef	SANBOOT_PROTO_AOE	/* AoE protocol */

/*
 * Boot connection vector support
 *
 */
#define	BCV			/* Boot Connection Vector entry point */
#define	INFOBLK			/* Use an INFOBLK structure to avoid protected mode */

/*
 * Boot connection protocols
 *
 */

#undef	BCV_PROTO_ISCSI		/* iSCSI protocol */

/*
 * Realmode support
 */
#if defined(BCV)
/* Only enable EBDA_ALLOCATE if BCV is defined, as EBDA memory is only used
 * when executing from the BCV entry point. */
#define        EBDA_ALLOCATE           /* Use EBDA allocation for lower memory */
#else /* BCV */
#undef EBDA_ALLOCATE
#endif /* BCV */
#undef	EBDA_RELOCATE		/* Move EBDA and use low memory just above it */
#define E820_MARKII		/* Report reserved regions in the E820 map */

/*
 * Name resolution modules
 *
 */

#define	DNS_RESOLVER		/* DNS resolver */
#undef	NMB_RESOLVER		/* NMB resolver */

/*
 * Image types
 *
 * Etherboot supports various image formats.  Select whichever ones
 * you want to use.
 *
 */
#undef	IMAGE_NBI		/* NBI image support */
#define	IMAGE_ELF		/* ELF image support */
#undef	IMAGE_FREEBSD		/* FreeBSD kernel image support */
#undef	IMAGE_MULTIBOOT		/* MultiBoot image support */
#undef	IMAGE_AOUT		/* a.out image support */
#undef	IMAGE_WINCE		/* WinCE image support */
#define	IMAGE_PXE		/* PXE image support */
#define	IMAGE_SCRIPT		/* gPXE script image support */
#define	IMAGE_BZIMAGE		/* Linux bzImage image support */
#define	IMAGE_COMBOOT		/* SYSLINUX COMBOOT image support */
#undef	IMAGE_EFI		/* EFI image support */

/*
 * User interface options
 *
 */
#define	SHELL_HIDE		/* Hide the shell */
#undef	SHELL_DISABLE		/* Disable all access to the shell */


/*
 * Command-line commands to include
 *
 */
#define	AUTOBOOT_CMD		/* Automatic booting */
#define	NVO_CMD			/* Non-volatile option storage commands */
#define	CONFIG_CMD		/* Option configuration console */
#define	IFMGMT_CMD		/* Interface management commands */
#define	ROUTE_CMD		/* Routing table management commands */
#define	IMAGE_CMD		/* Image management commands */
#define	DHCP_CMD		/* DHCP management commands */
#undef	SANBOOT_CMD		/* SAN boot commands */
#undef  MENU_CMD		/* Menu commands */
//#define	SFCONFIG_CMD		/* Solarflare configuration tool */
#undef	BOOT_SERV_CMD		/* Boot server discovery commands */
#undef	BOOTCONNECT_CMD		/* Boot connection command */
#define	MISC_CMD		/* Miscelanious commands */

#endif /* ! defined(EFI_BUILD) */

/*
 * ROM-specific options
 *
 */
#undef  NONPNP_HOOK_INT19       /* Hook INT19 on non-PnP BIOSes  */

/*
 * Obscure configuration options
 *
 * You probably don't need to touch these.
 *
 */

#undef	BUILD_SERIAL		/* Include an automatic build serial
				 * number.  Add "bs" to the list of
				 * make targets.  For example:
				 * "make bin/rtl8139.dsk bs" */
#undef	BUILD_ID		/* Include a custom build ID string,
				 * e.g "test-foo" */
#define	NULL_TRAP		/* Attempt to catch NULL function calls */
#undef	GDBSERIAL		/* Remote GDB debugging over serial */
#undef	GDBUDP			/* Remote GDB debugging over UDP
				 * (both may be set) */
#define	DEBUG_EXCEPTIONS	/* Debugging of protected mode exceptions
				 * (can't be used with GDB options) */

#undef	SFCONFIG_NOHELP		/* Remove tooltip help to save space */

#undef CONSOLE_SERIAL		/* Provide a serial console as well
				 * (can't be used with GDB options) */

#undef EFX_ENABLE_NVLOG		/* Log packets to NVRAM */
#define EFX_NVLOG_TCP_HDR_ONLY	/* For TCP packets only log headers to NVRAM */

/*
 * NOTE: certain options are set via CFLAGS_c in Makefile.housekeeping to avoid
 * adding dependencies on <config/general.h> in large numbers of files:
 *
 *  USE_ERRFILETAB		- automatically build error code -> file mappings
 *  DISABLE_APPLICATORS		-  Disable applicators, because sfc code consumes
 *				   settings from the correct context
 */

#endif /* CONFIG_GENERAL_H */
