/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include "crypto/axtls/crypto.h"
#include <gpxe/crypto.h>
#include <gpxe/sha1.h>

static void sha1_init ( void *ctx ) {
	SHA1Init ( ctx );
}

static void sha1_update ( void *ctx, const void *data, void *dst __unused,
			  size_t len ) {
	SHA1Update ( ctx, data, len );
}

static void sha1_final ( void *ctx, void *out ) {
	SHA1Final ( ctx, out );
}

struct crypto_algorithm sha1_algorithm = {
	.name		= "sha1",
	.ctxsize	= SHA1_CTX_SIZE,
	.blocksize	= 64,
	.digestsize	= SHA1_DIGEST_SIZE,
	.init		= sha1_init,
	.encode		= sha1_update,
	.final		= sha1_final,
};
