/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <gpxe/settings.h>
#include <gpxe/dhcpopts.h>
#include <gpxe/dhcp.h>
#include <gpxe/init.h>
#include <gpxe/sf_settings.h>

/** @file
 *
 * Default settings block for Solarflare NICs
 *
 */

static uint8_t default_settings_data[] = {
	DHCP_EB_ENCAP,
	DHCP_OPTION (
	    DHCP_ENCAPSULATED ( DHCP_EB_BOOT_IMAGE ),
	    DHCP_BYTE ( BOOTIMAGE_PXE_UEFI ),

	    DHCP_ENCAPSULATED ( DHCP_EB_LINKUP_DELAY ),
	    DHCP_BYTE ( 5 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_BANNER_DELAY ),
	    DHCP_BYTE ( 2 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_BOOTSKIP_DELAY ),
	    DHCP_BYTE ( 5 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_INITIATOR_DHCP ),
	    DHCP_BYTE ( 1 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_INITIATOR_IQN_DHCP ),
	    DHCP_BYTE ( 1 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_LUN_BUSY_RETRIES ),
	    DHCP_BYTE ( 2 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_TARGET_DHCP ),
	    DHCP_BYTE ( 1 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_MPIO_ATTEMPTS ),
	    DHCP_BYTE ( 3 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_MSIX_LIMIT ),
	    DHCP_WORD ( 32 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_VF_COUNT ),
	    DHCP_BYTE ( 127 ),

	    DHCP_ENCAPSULATED ( DHCP_EB_VF_MSIX_LIMIT ),
	    DHCP_BYTE ( 1 )
	),

	DHCP_RP_ENCAP,
	DHCP_OPTION (
	    DHCP_ENCAPSULATED ( DHCP_RP_PORT ),
	    DHCP_WORD ( 3260 ),

	    DHCP_ENCAPSULATED ( DHCP_RP_LUN ),
	    DHCP_BYTE ( 0 )
	),

	DHCP_END
};

static struct {
	struct generic_settings generic;
	struct refcnt refcounter;
	struct dhcp_options dhcpopts;
} sfc_defaults;

struct settings *sf_default_settings = &sfc_defaults.generic.settings;

static int sfc_defaults_fetch( struct settings *settings __unused,
			      struct setting *setting, void *data, size_t len)
{
	return dhcpopt_fetch ( &sfc_defaults.dhcpopts, setting->tag, data,
			       len );
}

static struct settings_operations sfc_defaults_settings_operations = {
	.fetch = sfc_defaults_fetch,
};

static void default_startup ( void ) {
	generic_settings_init ( &sfc_defaults.generic, &sfc_defaults.refcounter );
	dhcpopt_init ( &sfc_defaults.dhcpopts, default_settings_data,
	               sizeof ( default_settings_data ), dhcpopt_no_realloc );
	sfc_defaults.generic.settings.op = &sfc_defaults_settings_operations;
	register_settings ( sf_default_settings, NULL,
			    "sfc_defaults" );
}

static void default_shutdown(int booting __unused)
{
	static int done = 0;

	if (!done)
		unregister_settings(sf_default_settings);
	done++;
}

struct startup_fn default_settings_fn __startup_fn ( STARTUP_EARLY ) = {
	.startup = default_startup,
	.shutdown = default_shutdown
};
