/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef _GPXE_HP_OCSD_H
#define _GPXE_HP_OCSD_H

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdint.h>
#include <gpxe/smbios.h>

#define smbios_hp_ocsd_type 0xE5
#define ocsd_struct_version 2
#define option_card_field_version 1
#define ocsd_signature ('$' + (((uint32_t)'O') << 8 ) + \
			(((uint32_t)'C') << 16) + (((uint32_t)'S') << 24))
#define hdd_signature ('$' + (((uint32_t)'H') << 8 ) + \
		       (((uint32_t)'D') << 16) + (((uint32_t)'D') << 24))

/* Structure equivalent to definitions from
 * HP 'Option Card Sensor Data Reporting' v 1.4
 */
struct reserved_mem_entry {
	uint32_t signature;
	uint64_t physical_address;
	uint32_t size;
} __attribute__((packed));

struct smbios_reserved_mem_struct {
	struct smbios_header headers;
	struct reserved_mem_entry entries[0];
} __attribute__((packed));

struct option_card_fields {
	uint32_t version_number;
	uint16_t bus;
	uint16_t device;
	uint32_t reserved_1[3];
	uint32_t flags;
	uint32_t vendor_defined;
	uint32_t reserved_2[8];
	uint32_t checksum;

} __attribute__((packed));

enum sensor_type {
	no_sensor,
	thermal,
	board_power,
	chip_power
};

enum sensor_location {
	not_present,
	internal_asic,
	// "Near hottest part" is deprecated
	near_main_pci,
	near_rear,
	other,
	off_board
};

enum sensor_config {
	// Power readings.
	watts = 0,
	half_watts,
	// Temperature readings.
	signed_degrees_byte = 0,
	signed_half_degrees_byte,
	unsigned_degrees_byte,
	signed_degrees_32bit,
	signed_half_degrees_32bit,
	unsigned_degrees_32bit,
	relative
};

#define ocsd_sensor_enabled_bit 0
#define ocsd_sensor_enabled(status) (!!(status & (1 << ocsd_sensor_enabled_bit)))
#define ocsd_sensor_present_bit 1
#define ocsd_sensor_present(status) (!!(status & (1 << ocsd_sensor_present_bit)))
struct sensor_field {
	uint32_t type;
	uint32_t location;
	uint32_t caution_threshold;
	uint32_t max_continuous_threshold;
	uint16_t config;
	uint16_t status;
	uint32_t reading;
	uint32_t update_count;
	uint32_t checksum;
} __attribute__((packed));

struct card_entry {
	struct option_card_fields card_fields;
	struct sensor_field sensors[3];
} __attribute__((packed));

struct ocsd_buf {
	uint32_t version_number;
	uint32_t buffer_size;
	uint32_t max_card_count;
	uint32_t entry_size;
	uint32_t buffer_start;
	uint32_t reserved_1[3];
	uint32_t requested_update_interval;
	uint32_t reserved_2[5];
	uint32_t card_count;
	uint32_t checksum;
	/* We don't  use this field as the specification requires to use the
	 * fields 'option card buffer start' and 'option card & sensor size' to
	 * compute the card entry address.
	 */
	/* struct card_entry card_entries[0]; */
} __attribute__((packed));

#endif /* _GPXE_HP_OCSD_H */
